package at.ac.uibk.gitsearch.service;

import at.ac.uibk.gitsearch.domain.Review;
import at.ac.uibk.gitsearch.domain.ReviewRating;
import at.ac.uibk.gitsearch.domain.User;
import at.ac.uibk.gitsearch.domain.enumeration.ReviewStatus;
import at.ac.uibk.gitsearch.repository.jpa.ReviewRatingRepository;
import at.ac.uibk.gitsearch.repository.jpa.ReviewRatingRepository.StatusAndCount;
import at.ac.uibk.gitsearch.repository.jpa.ReviewRepository;
import at.ac.uibk.gitsearch.security.jwt.TokenProvider;
import at.ac.uibk.gitsearch.service.dto.ReviewDTO;
import at.ac.uibk.gitsearch.service.dto.ReviewRatingDTO;
import at.ac.uibk.gitsearch.service.dto.ReviewRequest;
import at.ac.uibk.gitsearch.service.dto.StatisticsDTO;
import at.ac.uibk.gitsearch.service.mapper.ReviewRatingMapper;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;
import javax.transaction.Transactional;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.core.userdetails.UsernameNotFoundException;
import org.springframework.stereotype.Service;

@Service
@Transactional
public class ReviewService {

    private static final String REVIEW_UPDATED = "Review Updated";

    private static final Logger LOGGER = LoggerFactory.getLogger(ReviewService.class);

    @Autowired
    private UserService userService;

    @Autowired
    private ReviewRepository reviewRepository;

    @Autowired
    private ReviewRatingRepository reviewRatingRepository;

    @Autowired
    private StatisticsService statisticsService;

    @Autowired
    protected TokenProvider tokenProvider;

    @Autowired
    private MailService mailService;

    @Autowired
    private ReviewRatingMapper reviewRatingMapper;

    private static final String emailStart = "The review for ";

    public Review createReview(ReviewRequest reviewRequest) throws UsernameNotFoundException {
        Review review = new Review();
        review.setResource(reviewRequest.getResource());
        review.setLink(reviewRequest.getLink());
        review.setResourceID(reviewRequest.getResourceID());
        User loggedInUser = userService.getUserWithAuthorities().orElse(null);
        if (loggedInUser == null) {
            throw new UsernameNotFoundException("User not found");
        }
        review.setRequestedBy(loggedInUser.getId());
        Set<ReviewRating> reviewRatings = new HashSet<>();

        reviewRequest
            .getUsers()
            .stream()
            .map(email -> userService.getUserByEmail(email).get())
            .forEach(user -> {
                ReviewRating reviewRating = new ReviewRating();
                reviewRating.setReview(review);
                reviewRating.setUser(user);
                reviewRating.setComment("");
                reviewRating.setStatus(ReviewStatus.REVIEW_IN_PROGRESS);
                reviewRatings.add(reviewRating);
                review.setComments(reviewRatings);
                reviewRepository.save(review);
                reviewRatingRepository.save(reviewRating);
                mailService.sendEmail(
                    user.getEmail(),
                    REVIEW_UPDATED,
                    "You have been selected to be a reviewer for " +
                    review.getResource() +
                    ". Please go to the sharing platform and review the resource.",
                    false,
                    false
                );
            });

        return review;
    }

    public void updateStatus(Long id, ReviewStatus status) {
        LOGGER.debug("REST request to update a review status");
        final Optional<Review> reviewO = reviewRepository.findById(id);
        if (reviewO.isEmpty()) {
            LOGGER.error("Cannot update nonexisting Review {}", id);
            return;
        }
        Review review = reviewO.get();
        review
            .getComments()
            .stream()
            .forEach(reviewRating -> {
                reviewRating.setStatus(status);
                reviewRatingRepository.save(reviewRating);
            });
    }

    public void updateReview(ReviewDTO review) {
        LOGGER.debug("REST request to update a review");
        final Optional<Review> reviewO = reviewRepository.findById(review.getId());
        if (reviewO.isEmpty()) {
            LOGGER.error("Cannot update nonexisting Review {}", review.getId());
            return;
        }
        Review reviewEntity = reviewO.get();
        reviewEntity
            .getComments()
            .stream()
            .filter(reviewRating -> !review.getUsers().contains(reviewRating.getUser().getEmail()))
            .forEach(reviewRating -> reviewRatingRepository.delete(reviewRating));
        reviewEntity.setResource(review.getResource());
        reviewEntity.setLink(review.getLink());

        Set<ReviewRating> reviewRatings = new HashSet<>();
        review
            .getUsers()
            .stream()
            .filter(email -> reviewEntity.getComments().stream().noneMatch(reviewRating -> reviewRating.getUser().getEmail().equals(email)))
            .forEach(email -> {
                ReviewRating reviewRating = new ReviewRating();
                reviewRating.setReview(reviewEntity);
                reviewRating.setUser(userService.getUserByEmail(email).get());
                reviewRating.setComment("");
                reviewRating.setStatus(ReviewStatus.REVIEW_IN_PROGRESS);
                reviewRatings.add(reviewRating);
                reviewRatingRepository.save(reviewRating);
            });

        reviewEntity.setComments(reviewRatings);
        reviewRepository.save(reviewEntity);

        notifyReviewers(
            review.getId(),
            "You have been selected to be a reviewer for " +
            review.getResource() +
            ". Please go to the sharing platform and review the resource."
        );
    }

    public List<ReviewDTO> getAllReviews() {
        LOGGER.debug("REST request to get all reviews");
        List<Review> reviews = reviewRepository.findAll();
        return reviews
            .stream()
            .map(review -> {
                ReviewDTO temp = reviewToDTO(review);
                temp.setLink(review.getLink());
                return temp;
            })
            .collect(Collectors.toList());
    }

    public List<ReviewDTO> getAllReviewsWithStatistics() {
        LOGGER.debug("REST request to get all reviews with statistics");
        List<Review> reviews = reviewRepository.findAll();
        return reviews
            .stream()
            .map(review -> {
                ReviewDTO temp = reviewToDTO(review);
                temp.setLink(review.getLink());
                StatisticsDTO stats = statisticsService.findOneByExerciseID(review.getResourceID()).orElse(null);
                if (stats == null) {
                    temp.setBadgeRewarded(false);
                } else {
                    temp.setBadgeRewarded(stats.getBadgeRewarded());
                }
                return temp;
            })
            .collect(Collectors.toList());
    }

    public List<ReviewDTO> getReviewsByUser() {
        LOGGER.debug("REST request to get all reviews by user");
        User user = userService.getUserWithAuthorities().orElse(null);
        if (user == null) {
            return new ArrayList<>();
        }
        List<ReviewRating> reviewRating = reviewRatingRepository.findAllByUserId(user.getId());

        List<ReviewDTO> reviewRatingDTOs = reviewRating
            .stream()
            .map(reviewRating1 -> {
                ReviewDTO reviewDTO = new ReviewDTO(
                    reviewRating1.getReview().getId(),
                    reviewRating1.getReview().getResource(),
                    reviewRating1.getReview().getLink(),
                    Arrays.asList(user.getEmail()),
                    Arrays.asList(reviewRating1.getStatus().name()),
                    Arrays.asList(reviewRating1.getComment()),
                    "ANNONYMIZED"
                );
                StatisticsDTO stats = statisticsService.findOneByExerciseID(reviewRating1.getReview().getResourceID()).orElse(null);
                if (stats == null) {
                    reviewDTO.setBadgeRewarded(false);
                } else {
                    reviewDTO.setBadgeRewarded(stats.getBadgeRewarded());
                }
                return reviewDTO;
            })
            .collect(Collectors.toList());

        List<Review> reviews = reviewRepository.findAllByRequestedBy(user.getId());
        List<ReviewDTO> reviewDTOs = reviews
            .stream()
            .map(review -> {
                ReviewDTO temp = reviewToDTO(review);
                temp.setLink(review.getLink());
                StatisticsDTO stats = statisticsService.findOneByExerciseID(review.getResourceID()).orElse(null);
                if (stats == null) {
                    temp.setBadgeRewarded(false);
                } else {
                    temp.setBadgeRewarded(stats.getBadgeRewarded());
                }
                return temp;
            })
            .collect(Collectors.toList());
        reviewDTOs
            .stream()
            .forEach(review -> {
                review.setRequestedBy(user.getEmail());
                review.setUsers(review.getUsers().stream().map(x -> "ANNONYMIZED").collect(Collectors.toList()));
            });
        reviewDTOs.addAll(reviewRatingDTOs);
        return reviewDTOs;
    }

    /**
     * calculates the review statistics by status for this user
     *
     * @param user
     * @return
     */
    public ReviewRatingDTO.ReviewStatisticsDTO getReviewStatusCountsByCurrentUser() {
        User user = userService.getUserWithAuthorities().orElse(null);
        if (user == null) {
            throw new UsernameNotFoundException("User not found, unable to update ReviewRating");
        }
        return getReviewStatusCountsByUser(user);
    }

    /**
     * calculates the review statistics by status for this user
     *
     * @param user
     * @return
     */
    public ReviewRatingDTO.ReviewStatisticsDTO getReviewStatusCountsByUser(User user) {
        List<StatusAndCount> counts = reviewRatingRepository.countReviewsForUserByStatus(user);
        ReviewRatingDTO.ReviewStatisticsDTO result = new ReviewRatingDTO.ReviewStatisticsDTO();
        counts.forEach(entry -> result.put(entry.getStatus(), entry.getStatusCount()));

        List<StatusAndCount> countsForRequestingUser = reviewRatingRepository.countReviewsForRequestingUserByStatus(user.getId());
        countsForRequestingUser.forEach(entry -> result.putByUser(entry.getStatus(), entry.getStatusCount()));

        Integer badgeCount = reviewRatingRepository.countBadgesForRequestingUser(user.getId());
        if (badgeCount != null) {
            result.setBadgesRewarded(badgeCount);
            result.setReviewsRequestedByUser(result.getReviewsRequestedByUser() - badgeCount);
        }
        return result;
    }

    public void updateReviewRatingByUser(ReviewDTO reviewDTO) throws IllegalAccessException {
        LOGGER.debug("REST request to update a review by user");
        User user = userService.getUserWithAuthorities().orElse(null);
        if (user == null) {
            throw new UsernameNotFoundException("User not found, unable to update ReviewRating");
        }
        LOGGER.debug("REST request to update a review");
        final Optional<Review> reviewO = reviewRepository.findById(reviewDTO.getId());
        if (reviewO.isEmpty()) {
            LOGGER.error("Cannot update nonexisting Review {}", reviewDTO.getId());
            return;
        }
        Review review = reviewO.get();
        if (review.getRequestedBy().equals(user.getId()) && !"DELETION REQUESTED".equals(reviewDTO.getStatus().get(0))) {
            throw new IllegalAccessException("User is not allowed to update ReviewRating");
        }
        if (review.getRequestedBy().equals(user.getId()) && "DELETION REQUESTED".equals(reviewDTO.getStatus().get(0))) {
            notifyReviewers(review.getId(), emailStart + review.getResource() + " has been deleted by the user.");
            deleteReview(review.getId());
            return;
        }

        ReviewRating reviewRating = review
            .getComments()
            .stream()
            .filter(reviewRating1 -> reviewRating1.getUser().getEmail().equals(user.getEmail()))
            .findFirst()
            .get();
        reviewRating.setComment(reviewDTO.getComments().get(0));
        reviewRating.setStatus(ReviewStatus.valueOf(reviewDTO.getStatus().get(0)));
        reviewRatingRepository.save(reviewRating);
        mailService.sendEmail(
            userService.getUserById(review.getRequestedBy()).get().getEmail(),
            "New feedback for resource",
            "A reviewer has just updated your review request for resource: " + review.getResource(),
            false,
            false
        );
    }

    public void deleteReview(Long id) {
        LOGGER.debug("REST request to delete review: {}", id);
        Review review = reviewRepository.findById(id).get();
        StatisticsDTO statistics = statisticsService.findOneByExerciseID(review.getResourceID()).orElse(null);
        if (statistics != null) {
            statistics.setBadgeRewarded(false);
            statisticsService.save(statistics);
        }
        notifyReviewers(review.getId(), emailStart + review.getResource() + " has been deleted.");
        notifyRequester(review.getId(), emailStart + review.getResource() + " has been deleted.");

        reviewRatingRepository.deleteByReviewId(id);
        reviewRepository.deleteById(id);
    }

    public ReviewDTO reviewToDTO(Review review) {
        ReviewDTO reviewDTO = new ReviewDTO();
        reviewDTO.setResource(review.getResource());
        reviewDTO.setUsers(
            review.getComments().stream().map(reviewRating -> reviewRating.getUser().getEmail()).collect(Collectors.toList())
        );
        reviewDTO.setLink(review.getLink());
        reviewDTO.setRequestedBy(userService.getUserById(review.getRequestedBy()).get().getEmail());
        reviewDTO.setId(review.getId());
        reviewDTO.setStatus(
            review.getComments().stream().map(reviewRating -> reviewRating.getStatus().name()).collect(Collectors.toList())
        );
        reviewDTO.setComments(review.getComments().stream().map(reviewRating -> reviewRating.getComment()).collect(Collectors.toList()));
        reviewDTO.setResourceID(review.getResourceID());
        return reviewDTO;
    }

    public ReviewDTO getReview(Long id) {
        Review review = reviewRepository.findById(id).get();
        return reviewToDTO(review);
    }

    public void rewardBadge(Long id) {
        Review review = reviewRepository.findById(id).get();
        updateStatus(review.getId(), ReviewStatus.REVIEW_COMPLETED);
        statisticsService.rewardBadgeForStatisticsByExerciseID(review.getResourceID());
        notifyReviewers(
            review.getId(),
            emailStart + review.getResource() + " has been completed and a badge is now being displayed when people search for it."
        );
        notifyRequester(
            review.getId(),
            emailStart + review.getResource() + " has been completed and a badge is now being displayed when people search for it."
        );
    }

    public void requestReviewForExercise(ReviewRequest request) {
        String id = request.getResource();
        if (reviewRepository.findOneByResource(id).isPresent()) {
            LOGGER.warn("Review already exists for exercise: {}", id);
            return;
        }

        Review review = new Review();
        review.setResource(id);
        review.setLink(request.getLink());
        review.setResourceID(request.getResourceID());

        Optional<User> user = userService.getUserWithAuthorities();
        if (user.isPresent()) {
            review.setRequestedBy(user.get().getId());
            // reviewRating.setUser(user.get());
        } else {
            LOGGER.error("User not found");
            return;
        }
        Set<ReviewRating> reviewRatings = new HashSet<>();
        review.setComments(reviewRatings);
        reviewRepository.save(review);

        mailService.sendEmail(
            System.getenv("REVIEW_ADMIN"),
            "NEW REVIEW REQUESTED",
            user.get().getEmail() + " requested a review for " + request.getResource(),
            false,
            false
        );
    }

    public void notifyReviewers(Long id, String message) {
        Review review = reviewRepository.findById(id).get();
        List<String> emails = review
            .getComments()
            .stream()
            .map(reviewRating -> reviewRating.getUser().getEmail())
            .collect(Collectors.toList());
        emails.stream().forEach(email -> mailService.sendEmail(email, REVIEW_UPDATED, message, false, false));
    }

    // notify requester of review
    public void notifyRequester(Long id, String message) {
        Review review = reviewRepository.findById(id).get();
        User user = userService.getUserById(review.getRequestedBy()).get();
        mailService.sendEmail(user.getEmail(), REVIEW_UPDATED, message, false, false);
    }

    public List<ReviewRatingDTO> getAllReviewRatings() {
        return reviewRatingRepository.findAll().stream().map(reviewRatingMapper::toDto).collect(Collectors.toList());
    }
}
